--
-- AutoLoadWoodCustom for FS25
-- Based on AutoLoadWood by Marhu, kenny456
-- @author: BMP Modified for configurable nodes in XML
-- @history: v1.0 - 2025-05-15 - Initial version with configurable nodes
AutoLoadWoodCustom = {};
AutoLoadWoodCustom.confDir = getUserProfileAppPath().. "modSettings/AutoLoadWoodCustom/";
local modName = g_currentModName

function AutoLoadWoodCustom:getTexts(i, text)
	if g_currentMission.AutoLoadWoodCustom == nil then
		g_currentMission.AutoLoadWoodCustom = {}
		g_currentMission.AutoLoadWoodCustom.texts = {}
	end
	g_currentMission.AutoLoadWoodCustom.texts[i] = text
end

function AutoLoadWoodCustom.prerequisitesPresent(specializations)
    return true
end

function AutoLoadWoodCustom.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "dynamicMountTriggerCallback", AutoLoadWoodCustom.dynamicMountTriggerCallback)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsTurnedOn", AutoLoadWoodCustom.getIsTurnedOn)
end

function AutoLoadWoodCustom.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "toggleState", AutoLoadWoodCustom.toggleState)
	SpecializationUtil.registerFunction(vehicleType, "toggleLoad", AutoLoadWoodCustom.toggleLoad)
	SpecializationUtil.registerFunction(vehicleType, "toggleUnload", AutoLoadWoodCustom.toggleUnload)
	SpecializationUtil.registerFunction(vehicleType, "updateValues", AutoLoadWoodCustom.updateValues)
	SpecializationUtil.registerFunction(vehicleType, "updateLoad", AutoLoadWoodCustom.updateLoad)
	SpecializationUtil.registerFunction(vehicleType, "updateUnload", AutoLoadWoodCustom.updateUnload)
	SpecializationUtil.registerFunction(vehicleType, "saveToXml", AutoLoadWoodCustom.saveToXml)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXml", AutoLoadWoodCustom.loadFromXml)
end

function AutoLoadWoodCustom.registerEventListeners(vehicleType)
	for _, n in pairs({"onLoad", "onPostLoad", "saveToXMLFile", "onUpdate", "onDraw", "onRegisterActionEvents", "saveToXml", "loadFromXml", "onReadStream", "onWriteStream"}) do
		SpecializationUtil.registerEventListener(vehicleType, n, AutoLoadWoodCustom)
	end
end

function AutoLoadWoodCustom:onLoad(savegame)
    self.spec_autoLoadWoodCustom = {}
    local spec = self.spec_autoLoadWoodCustom
    local xmlFile = self.xmlFile
    spec.event_IDs = {}
    
    spec.searchSizeY = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.AutoLoadWood#SizeY"), 5)
    spec.searchSizeZ = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.AutoLoadWood#SizeX"), 10)
    spec.autoLoadSpeed = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.AutoLoadWood#Speed"), 500)
    spec.maxPileHeight = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.AutoLoadWood#maxPileHeight"), 5.0)
    spec.showNotifications = Utils.getNoNil(getXMLBool(xmlFile.handle, "vehicle.AutoLoadWood#showNotifications"), true)
    spec.maxLogsPerRow = Utils.getNoNil(getXMLInt(xmlFile.handle, "vehicle.AutoLoadWood#maxLogsPerRow"), 4) -- Максимум бревен в ряду
    spec.autoLoadElaps = 0
    spec.logsInRow = 0 -- Количество бревен в текущем ряду
    spec.currentPileHeight = 0 -- Текущая высота штабеля
    spec.rowHeight = 0.5 -- Высота ряда (будет обновляться при загрузке первого бревна)
    
    local dynamicMountTriggerNode = getXMLString(xmlFile.handle, "vehicle.AutoLoadWood#dynamicMountTriggerNode")
    local autoLoadWoodNode = getXMLString(xmlFile.handle, "vehicle.AutoLoadWood#autoLoadWoodNode")
    local loadMarkerNode = getXMLString(xmlFile.handle, "vehicle.loadMarker#node")
    local unloadMarkerNode = getXMLString(xmlFile.handle, "vehicle.unloadMarker#node")
    
    spec.dynamicMountTrigger = dynamicMountTriggerNode and I3DUtil.indexToObject(self.components, dynamicMountTriggerNode) or nil
    spec.autoLoadWood = autoLoadWoodNode and I3DUtil.indexToObject(self.components, autoLoadWoodNode) or nil
    spec.loadMarker = loadMarkerNode and I3DUtil.indexToObject(self.components, loadMarkerNode) or nil
    spec.unloadMarker = unloadMarkerNode and I3DUtil.indexToObject(self.components, unloadMarkerNode) or nil
    
    if spec.dynamicMountTrigger then
        addTrigger(spec.dynamicMountTrigger, "dynamicMountTriggerCallback", self)
    else
        print("Error: dynamicMountTrigger is nil, check dynamicMountTriggerNode in XML")
    end
    
    local x, y, z
    if spec.dynamicMountTrigger then
        x, y, z = getTranslation(spec.dynamicMountTrigger)
    else
        x, y, z = 0, 0, 0
    end
    
    local key = "vehicle.AutoLoadWood.fillPose"
    local fillPoseXOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#xOff"), 0)
    local fillPoseYOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#yOff"), 2)
    local fillPoseZOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#zOff"), 0)
    
    local key = "vehicle.AutoLoadWood.unloadPose"
    local unloadPoseXOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#xOff"), 0)
    local unloadPoseYOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#yOff"), 5)
    local unloadPoseZOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#zOff"), 0)
        
    spec.fillPoseNodeX = x + fillPoseXOff
    spec.fillPoseNodeY = 1.4 + fillPoseYOff
    spec.fillPoseNodeZ = z + fillPoseZOff
    spec.fillPoseRoot = createTransformGroup("fillPoseRoot")
    spec.fillPoseNode = createTransformGroup("fillPoseNode")
    if spec.autoLoadWood then
        link(spec.autoLoadWood, spec.fillPoseRoot)
        link(spec.fillPoseRoot, spec.fillPoseNode)
        setTranslation(spec.fillPoseRoot, spec.fillPoseNodeX, spec.fillPoseNodeY, spec.fillPoseNodeZ)
        setRotation(spec.fillPoseRoot, math.rad(-90), 0, 0)
    else
        print("Error: autoLoadWood is nil, check autoLoadWoodNode in XML")
    end
    
    local linkNode = spec.dynamicMountTrigger and getParent(getParent(spec.dynamicMountTrigger)) or self.rootNode
    spec.unloadPoseRoot = { createTransformGroup("unloadPoseRoot1"), createTransformGroup("unloadPoseRoot2") }
    spec.unloadPoseNode = { createTransformGroup("unloadPoseNode1"), createTransformGroup("unloadPoseNode2") }
    spec.unloadPoseRootX = 2
    spec.unloadPoseRootY = 0
    spec.unloadPoseRootZ = 0
    for i = 1, 2 do
        link(linkNode, spec.unloadPoseRoot[i])
        setTranslation(spec.unloadPoseRoot[i], spec.unloadPoseRootX, spec.unloadPoseRootY, spec.unloadPoseRootZ)
        setRotation(spec.unloadPoseRoot[i], math.rad(0), math.rad(0), 0)
        spec.unloadPoseRootX = spec.unloadPoseRootX * -1
    end
    
    spec.unloadPoseNodeX = unloadPoseXOff
    spec.unloadPoseNodeY = unloadPoseYOff
    spec.unloadPoseNodeZ = unloadPoseZOff
    for i = 1, 2 do                    
        link(spec.unloadPoseRoot[i], spec.unloadPoseNode[i])
        setTranslation(spec.unloadPoseNode[i], spec.unloadPoseNodeX, spec.unloadPoseNodeY, spec.unloadPoseNodeZ)
        setRotation(spec.unloadPoseNode[i], math.rad(-90), 0, 0)
        spec.unloadPoseNodeX = spec.unloadPoseNodeX * -1
    end
    
    if spec.loadMarker then
        spec.markerX, spec.markerY, spec.markerZ = getTranslation(spec.loadMarker)
        setVisibility(spec.loadMarker, false)
        link(spec.fillPoseRoot, spec.loadMarker)
        setRotation(spec.loadMarker, math.rad(90), 0, math.rad(90))
        setTranslation(spec.loadMarker, 0, 0, -3.1)
    end
    if spec.unloadMarker then
        spec.markerX, spec.markerY, spec.markerZ = getTranslation(spec.unloadMarker)
        setVisibility(spec.unloadMarker, false)
        link(spec.unloadPoseRoot[1], spec.unloadMarker)
    end
	
    if self.isServer then
        local key = "vehicle.AutoLoadWood.woodTrigger"
        local woodTriggerXOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#xOff"), 1.2)
        local woodTriggerYOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#yOff"), -3)
        local woodTriggerZOff = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#zOff"), 1.5)

        local key = "vehicle.AutoLoadWood.limitToLen"
        spec.loadMinLen = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#minLen"), 2.5)
        spec.loadMaxLen = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#maxLen"), 20)
        spec.loadMaxAtt = Utils.getNoNil(getXMLInt(xmlFile.handle, key .. "#delimbOk"), 0)
		
        local linkNode = spec.dynamicMountTrigger and getParent(spec.dynamicMountTrigger) or self.rootNode
        spec.woodTrigger = {
            createTransformGroup("woodTrigger1"),
            createTransformGroup("woodTrigger2"),
            createTransformGroup("woodTrigger3"),
            createTransformGroup("woodTrigger4")
        }
		
        local rot = math.rad(90)
        for i = 1, 4 do					
            link(linkNode, spec.woodTrigger[i])
            setTranslation(spec.woodTrigger[i], x + woodTriggerXOff, y + woodTriggerYOff, z + woodTriggerZOff)
            setRotation(spec.woodTrigger[i], 0, rot, 0)
            woodTriggerZOff = woodTriggerZOff * -1
            if i == 2 then 
                woodTriggerXOff = woodTriggerXOff * -1
                rot = math.rad(-90)
            end
        end
    end
    self.onDeleteBackup = self.delete
    self.delete = Utils.prependedFunction(self.delete, AutoLoadWoodCustom.onDeleteAL)
end

function AutoLoadWoodCustom:onPostLoad(savegame)
    local spec = self.spec_autoLoadWoodCustom
    local xmlFile = self.xmlFile
    if spec.event_IDs == nil then
        spec.event_IDs = {}
    end
    spec.woodLoad = {}
    spec.autoload = false
    spec.rightSide = false
    spec.unload = false
    spec.showLoadPos = false
    spec.numPiles = 1 -- Начинаем с 1 штабеля
    spec.pile = 1
    spec.showUnloadPos = false
    spec.unloadPos = spec.unloadPoseRootX
    spec.unloadScale = 1
    spec.unloadRot = false
    spec.trailer = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.base.size#length"), 15)
    spec.space = 0.3
    spec.maxLen = Utils.getNoNil(getXMLFloat(xmlFile.handle, "vehicle.AutoLoadWood.limitToLen#maxLen"), spec.trailer) -- Чтение maxLen
    spec.maxPiles = getXMLInt(xmlFile.handle, "vehicle.AutoLoadWood#maxPiles") -- Чтение maxPiles из XML
    spec.rootShiftRotX = 6
    spec.rootShiftRotXMax = 9
    AutoLoadWoodCustom.showHelp = true
    spec.logsCount = 0
    spec.logsMass = 0
    local x, y, z = getCenterOfMass(self.rootNode)
    spec.massYOrig = y
    spec.massYLast = spec.massYOrig
    spec.mass = 0
    spec.count = 0
    spec.firstRun = true
    
    print("Debug: onPostLoad, trailer = " .. tostring(spec.trailer) .. ", maxLen = " .. tostring(spec.maxLen) .. ", maxPiles = " .. tostring(spec.maxPiles) .. ", maxLength = " .. tostring(spec.maxLength))
    self:updateValues()
    print("Debug: after updateValues, maxPiles = " .. tostring(spec.maxPiles) .. ", maxLength = " .. tostring(spec.maxLength))
    self:updateLoad()
    self:updateUnload()
    
    if savegame ~= nil then
        local xmlFile = savegame.xmlFile
        local key = savegame.key .. "." .. modName .. ".AutoLoadWoodCustom"
        local rightSide = Utils.getNoNil(getXMLBool(xmlFile.handle, key .. "#rightSide"), spec.rightSide)
        local showLoadPos = Utils.getNoNil(getXMLBool(xmlFile.handle, key .. "#showLoadPos"), spec.showLoadPos)
        local showUnloadPos = Utils.getNoNil(getXMLBool(xmlFile.handle, key .. "#showUnloadPos"), spec.showUnloadPos)
        local numPiles = Utils.getNoNil(getXMLInt(xmlFile.handle, key .. "#numPiles"), spec.numPiles)
        local pile = Utils.getNoNil(getXMLInt(xmlFile.handle, key .. "#pile"), spec.pile)
        local unloadPos = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#unloadPos"), spec.unloadPos)
        local unloadScale = Utils.getNoNil(getXMLFloat(xmlFile.handle, key .. "#unloadScale"), spec.unloadScale)
        local unloadRot = Utils.getNoNil(getXMLBool(xmlFile.handle, key .. "#unloadRot"), spec.unloadRot)
        
        self:toggleState(spec.autoload, rightSide, spec.unload, true)
        self:toggleLoad(showLoadPos, numPiles, pile, true)
        self:toggleUnload(showUnloadPos, unloadPos, unloadScale, unloadRot, true)
    end
    
    local configFile = AutoLoadWoodCustom.confDir .. "AutoLoadWoodCustomConfig.xml"
    if fileExists(configFile) then
        AutoLoadWoodCustom.configXml = loadXMLFile("AutoLoadWoodCustom_XML", configFile)
        self:loadFromXml(self:getFullName())
    else
        createFolder(getUserProfileAppPath() .. "modSettings/")
        createFolder(AutoLoadWoodCustom.confDir)
        AutoLoadWoodCustom.configXml = createXMLFile("AutoLoadWoodCustom_XML", configFile, "AutoLoadWoodCustomConfig")
        self:saveToXml()
        saveXMLFile(AutoLoadWoodCustom.configXml)
    end
    if self.spec_vehicleStraps ~= nil then
        self.spec_vehicleStraps.modInitialized = false
        self.spec_vehicleStraps.modAllowed = false
        self.dynamicMountTriggerCallback = Utils.overwrittenFunction(self.dynamicMountTriggerCallback, AutoLoadWoodCustom.dynamicMountTriggerCallback)
    end
end

function AutoLoadWoodCustom:onDelete()
	local spec = self.spec_autoLoadWoodCustom
	if spec.dynamicMountTrigger ~= nil then
		removeTrigger(spec.dynamicMountTrigger)
	end
end

function AutoLoadWoodCustom:onDeleteAL()
	local spec = self.spec_autoLoadWoodCustom
	if spec.dynamicMountTrigger ~= nil then
		removeTrigger(spec.dynamicMountTrigger)
	end
end

function AutoLoadWoodCustom:saveToXMLFile(xmlFile, key)
	local spec = self.spec_autoLoadWoodCustom
	setXMLBool(xmlFile.handle, key .. "#rightSide", spec.rightSide)
	setXMLBool(xmlFile.handle, key .. "#showLoadPos", spec.showLoadPos)
	setXMLBool(xmlFile.handle, key .. "#showUnloadPos", spec.showUnloadPos)
	setXMLInt(xmlFile.handle, key .. "#numPiles", spec.numPiles)
	setXMLInt(xmlFile.handle, key .. "#pile", spec.pile)
	setXMLFloat(xmlFile.handle, key .. "#unloadPos", spec.unloadPos)
	setXMLFloat(xmlFile.handle, key .. "#unloadScale", spec.unloadScale)
	setXMLBool(xmlFile.handle, key .. "#unloadRot", spec.unloadRot)
end

function AutoLoadWoodCustom:loadFromXml()
	local spec = self.spec_autoLoadWoodCustom
	if AutoLoadWoodCustom.configXml == nil then
		return
	end
	if getXMLBool(AutoLoadWoodCustom.configXml, "AutoLoadWoodCustomConfig.showHelp") ~= nil then
		AutoLoadWoodCustom.showHelp = getXMLBool(AutoLoadWoodCustom.configXml, "AutoLoadWoodCustomConfig.showHelp")
	end
end

function AutoLoadWoodCustom:saveToXml()
	local spec = self.spec_autoLoadWoodCustom
	if AutoLoadWoodCustom.configXml == nil then
		return
	end
	setXMLBool(AutoLoadWoodCustom.configXml, "AutoLoadWoodCustomConfig.showHelp", AutoLoadWoodCustom.showHelp)
	saveXMLFile(AutoLoadWoodCustom.configXml)
end

function AutoLoadWoodCustom:onUpdate(dt, vehicle)
    local spec = self.spec_autoLoadWoodCustom
    local areBeltsFasten = false
    if self.spec_tensionBelts ~= nil then
        areBeltsFasten = self.spec_tensionBelts.fastenedAllBeltsState
    end
    if self:getIsActive() and (not areBeltsFasten or spec.firstRun) then
        spec.firstRun = false
        spec.count = 0
        spec.mass = 0
        for k, shape in pairs(spec.woodLoad) do
            if entityExists(shape) then
                spec.count = spec.count + 1
                spec.mass = spec.mass + getMass(shape)
            else
                spec.woodLoad[k] = nil
            end
        end
    end
    local massY = spec.massYOrig
    if spec.count > 0 and not areBeltsFasten then
        local k = -0.425 * (spec.trailer or 15)
        if k ~= 0 then
            massY = (spec.mass / k) + spec.massYOrig
        else
            print("Warning: k is zero, skipping massY calculation to avoid division by zero")
        end
    end
    if self.lastSpeed < 0.011 then
        massY = spec.massYOrig
    end
    if massY ~= spec.massYLast then
        spec.massYLast = massY
        local x, y, z = getCenterOfMass(self.rootNode)
        setCenterOfMass(self.rootNode, x, massY, z)
    end
    if spec.autoload then
        spec.autoLoadElaps = spec.autoLoadElaps + dt
        if spec.autoLoadElaps >= spec.autoLoadSpeed then
            spec.autoLoadElaps = 0
            if self.isServer then
                if entityExists(spec.lastWoodShapes) and spec.woodLoad[spec.lastWoodShapes] == nil then
                    if spec.pile + 1 <= spec.numPiles then
                        spec.pile = spec.pile + 1
                        spec.lastWoodShapes = 0
                        spec.currentPileHeight = 0
                        spec.logsInRow = 0
                        self:toggleLoad(spec.showLoadPos, spec.numPiles, spec.pile)
                    else
                        self:toggleState(false, spec.rightSide, spec.unload)
                        return
                    end
                end
                local triggerNum = (spec.rightSide and 2 or 0) + 1
                local x, y, z = getWorldTranslation(spec.woodTrigger[triggerNum])
                local nx, ny, nz = localDirectionToWorld(spec.woodTrigger[triggerNum], 1, 0, 0)
                local yx, yy, yz = localDirectionToWorld(spec.woodTrigger[triggerNum], 0, 1, 0)
                local shape, minY, maxY, minZ, maxZ = findSplitShape(x, y, z, nx, ny, nz, yx, yy, yz, spec.searchSizeY, spec.searchSizeZ)
                if shape == nil or not entityExists(shape) or getRigidBodyType(shape) ~= 2 or spec.badTree == shape then
                    triggerNum = triggerNum + 1
                    x, y, z = getWorldTranslation(spec.woodTrigger[triggerNum])
                    nx, ny, nz = localDirectionToWorld(spec.woodTrigger[triggerNum], 1, 0, 0)
                    yx, yy, yz = localDirectionToWorld(spec.woodTrigger[triggerNum], 0, 1, 0)
                    shape, minY, maxY, minZ, maxZ = findSplitShape(x, y, z, nx, ny, nz, yx, yy, yz, spec.searchSizeY, spec.searchSizeZ)
                end
                local fX, fY, fZ = getTranslation(spec.fillPoseNode)
                local rowCompleted = false
                if fX < 0.5 then
                    fX = fX + 0.5
                else
                    fX = -0.8
                    rotate(spec.fillPoseNode, 0, 0, math.rad(180))
                    spec.fXRot = not spec.fXRot
                    rowCompleted = true
                end
                setTranslation(spec.fillPoseNode, fX, fY, fZ)
                if shape ~= nil and entityExists(shape) and getRigidBodyType(shape) == 2 then
                    local sizeX, sizeY, sizeZ, numConvexes, numAttachments = getSplitShapeStats(shape)
                    local maxSize = math.max(sizeX, math.max(sizeY, sizeZ))
                    local a, b, c = localToWorld(spec.woodTrigger[triggerNum], 0, (minY + maxY) * 0.5, (minZ + maxZ) * 0.5)
                    local sx, sy, sz = getWorldTranslation(shape)
                    local dist = MathUtil.vector3Length(a - sx, b - sy, c - sz)
                    local nx, ny, nz = localDirectionToWorld(shape, 0, 1, 0)
                    local lenBelow, lenAbove = getSplitShapePlaneExtents(shape, a, b, c, nx, ny, nz)
                    dist = dist - lenBelow
                    local diffX = (spec.maxLength - sizeX) * 0.5 * (spec.fXRot and -1 or 1)
                    local wx, wy, wz = localToWorld(spec.fillPoseNode, 0, -dist - (sizeX / 2) - diffX, 0)
                    local dx, dy, dz = worldToLocal(getParent(shape), wx, wy, wz)
                    local nx, ny, nz = localDirectionToWorld(spec.fillPoseNode, 1, 0, 0)
                    local yx, yy, yz = localDirectionToWorld(spec.fillPoseNode, 0, 1, 0)
                    local x, y, z = worldDirectionToLocal(getParent(shape), nx, ny, nz)
                    local xdir, ydir, zdir = worldDirectionToLocal(getParent(shape), yx, yy, yz)
                    if numAttachments == nil then numAttachments = 0 end
                    if maxSize <= spec.loadMaxLen and maxSize >= spec.loadMinLen and numAttachments <= spec.loadMaxAtt then
                        -- Проверка высоты перед загрузкой
                        local newHeight = spec.currentPileHeight
                        if rowCompleted then
                            newHeight = spec.currentPileHeight + spec.rowHeight
                        end
                        if newHeight <= spec.maxPileHeight then
                            removeFromPhysics(shape)
                            setTranslation(shape, dx, dy, dz)
                            setDirection(shape, x, y, z, xdir, ydir, zdir)
                            addToPhysics(shape)
                            spec.lastWoodShapes = shape
                            spec.logsInRow = spec.logsInRow + 1
                            -- Устанавливаем высоту ряда на основе первого бревна в ряду
                            if spec.logsInRow == 1 then
                                spec.rowHeight = sizeY
                            end
                            -- Если ряд завершен, обновляем высоту штабеля
                            if rowCompleted then
                                spec.currentPileHeight = spec.currentPileHeight + spec.rowHeight
                                spec.logsInRow = 0
                                print("Debug: Row completed, currentPileHeight = " .. spec.currentPileHeight .. ", rowHeight = " .. spec.rowHeight)
                            end
                            print("Debug: Loaded log, sizeY = " .. sizeY .. ", logsInRow = " .. spec.logsInRow .. ", currentPileHeight = " .. spec.currentPileHeight)
                        else
                            print("Debug: Max pile height reached (" .. spec.maxPileHeight .. "m), stopping autoload")
                            if spec.showNotifications then
                                local message = string.format(g_i18n:getText("AUTOLOADWOODEW_MAX_HEIGHT_REACHED"), spec.maxPileHeight)
                                g_currentMission:showBlinkingWarning(message, 3000)
                            end
                            self:toggleState(false, spec.rightSide, spec.unload)
                        end
                    else
                        spec.badTree = shape
                    end
                end
            end
        end
    elseif spec.unload then
        spec.autoLoadElaps = spec.autoLoadElaps + dt
        if spec.autoLoadElaps >= spec.autoLoadSpeed then
            spec.autoLoadElaps = 0
            if self.isServer then
                local side = spec.rightSide and 2 or 1
                local x, y, z = getWorldTranslation(spec.fillPoseNode)
                local uX, uY, uZ = getTranslation(spec.unloadPoseNode[side])
                local findWood = false
                for k, shape in pairs(spec.woodLoad) do
                    if entityExists(shape) then
                        spec.shape = shape
                        local sx, sy, sz = getWorldTranslation(shape)
                        local dist = MathUtil.vector2Length(x - sx, z - sz)
                        uX = math.abs(uX)
                        if uX < spec.unloadPoseNodeX + spec.unloadShiftMax then
                            uX = uX + 0.6
                        else
                            uX = spec.unloadPoseNodeX
                        end
                        setTranslation(spec.unloadPoseNode[side], (spec.rightSide and -1 or 1) * uX, uY, uZ)
                        if getChild(shape, 'logCenter') == 0 then
                            local x, y, z = getWorldTranslation(shape)
                            local nx, ny, nz = localDirectionToWorld(shape, 0, 1, 0)
                            local lenBelow, lenAbove = getSplitShapePlaneExtents(shape, x, y, z, nx, ny, nz)
                            local sizeX, sizeY, sizeZ = getSplitShapeStats(shape)
                            local logCenter = createTransformGroup("logCenter")
                            link(shape, logCenter)
                            setTranslation(logCenter, 0, -lenBelow + (sizeX / 2), 0)
                            spec.logCenter = logCenter
                        else
                            local logCenter = getChild(shape, 'logCenter')
                            spec.logCenter = logCenter
                        end
                        if dist < 100 then
                            local wx, wy, wz = localToWorld(spec.unloadPoseNode[side], 0, -dist, 0)
                            local dx, dy, dz = worldToLocal(getParent(shape), wx, wy, wz)
                            local nx, ny, nz = localDirectionToWorld(spec.unloadPoseNode[side], 1, 0, 0)
                            local yx, yy, yz = localDirectionToWorld(spec.unloadPoseNode[side], 0, 1, 0)
                            local x, y, z = worldDirectionToLocal(getParent(shape), nx, ny, nz)
                            local xdir, ydir, zdir = worldDirectionToLocal(getParent(shape), yx, yy, yz)
                            removeFromPhysics(shape)
                            setTranslation(shape, dx, dy, dz)
                            setDirection(shape, x, y, z, xdir, ydir, zdir)
                            addToPhysics(shape)
                            local x, y, z = getWorldTranslation(spec.logCenter)
                            local a, b, c = getWorldTranslation(spec.unloadPoseNode[side])
                            local dist2 = MathUtil.vector2Length(x - a, z - c)
                            if dist2 > 0 then
                                local wx, wy, wz = localToWorld(spec.unloadPoseNode[side], 0, -dist + dist2, 0)
                                local dx, dy, dz = worldToLocal(getParent(shape), wx, wy, wz)
                                setTranslation(shape, dx, dy, dz)
                            end
                            local x, y, z = getWorldTranslation(spec.logCenter)
                            local a, b, c = getWorldTranslation(spec.unloadPoseNode[side])
                            local dist3 = MathUtil.vector2Length(x - a, z - c)
                            if dist3 > 0 then
                                local wx, wy, wz = localToWorld(spec.unloadPoseNode[side], 0, -dist + dist2 - dist3, 0)
                                local dx, dy, dz = worldToLocal(getParent(shape), wx, wy, wz)
                                setTranslation(shape, dx, dy, dz)
                            end
                            self:dynamicMountTriggerCallback(spec.dynamicMountTrigger, shape, false, true, false, shape)
                            if self.spec_timberTrailerInfo ~= nil and self.spec_timberTrailerInfo.logTrigger ~= nil then
                                self:logTriggerCallback(self.spec_timberTrailerInfo.logTrigger, shape, false, true, false, shape)
                            end
                            findWood = true
                            break
                        else
                            spec.woodLoad[k] = nil
                        end
                    else
                        spec.woodLoad[k] = nil
                    end
                end
                if findWood == false then
                    self:toggleState(spec.autoload, spec.rightSide, false)
                    setTranslation(spec.unloadPoseNode[side], spec.unloadPoseNodeX, spec.unloadPoseNodeY, spec.unloadPoseNodeZ)
                    spec.currentPileHeight = 0
                    spec.logsInRow = 0
                end
            end
        end
    end
    if self:getIsActiveForInput() then
        if spec.event_IDs ~= nil and g_dedicatedServerInfo == nil and self.isClient then
            for actionName, eventID in pairs(spec.event_IDs) do
                g_inputBinding:setActionEventActive(eventID, true)
                if actionName == InputAction.AUTOLOADWOODEW_TOGGLE_AUTO then
                    g_inputBinding:setActionEventText(eventID, spec.autoload and g_i18n:getText('AUTOLOADWOODEW_STOP_AUTOLOAD') or g_i18n:getText('AUTOLOADWOODEW_START_AUTOLOAD'))
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_UNLOAD then
                    g_inputBinding:setActionEventText(eventID, spec.unload and g_i18n:getText('AUTOLOADWOODEW_STOP_UNLOAD') or g_i18n:getText('AUTOLOADWOODEW_START_UNLOAD'))
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_SIDE then
                    g_inputBinding:setActionEventText(eventID, spec.rightSide and g_i18n:getText('AUTOLOADWOODEW_SWITCH_LEFT') or g_i18n:getText('AUTOLOADWOODEW_SWITCH_RIGHT'))
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_LOAD_MARKER then
                    g_inputBinding:setActionEventText(eventID, spec.showLoadPos and g_i18n:getText('AUTOLOADWOODEW_HIDE_LOAD_MARKER') or g_i18n:getText('AUTOLOADWOODEW_SHOW_LOAD_MARKER'))
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_UNLOAD_MARKER then
                    g_inputBinding:setActionEventText(eventID, spec.showUnloadPos and g_i18n:getText('AUTOLOADWOODEW_HIDE_UNLOAD_MARKER') or g_i18n:getText('AUTOLOADWOODEW_SHOW_UNLOAD_MARKER'))
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_HELP then
                    g_inputBinding:setActionEventText(eventID, AutoLoadWoodCustom.showHelp and g_i18n:getText('AUTOLOADWOODEW_HIDE_HELP') or g_i18n:getText('AUTOLOADWOODEW_SHOW_HELP'))
                    g_inputBinding:setActionEventTextVisibility(eventID, true)
                elseif actionName == InputAction.AUTOLOADWOODEW_TOGGLE_PILES then
                    local maxLength = spec.maxLength or 15
                    g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_AUTOLOADWOODEW_TOGGLE_PILES') .. ' : ' .. spec.numPiles .. ' (max. ' .. maxLength .. ' m)')
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_SWITCH_PILE then
                    g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_AUTOLOADWOODEW_SWITCH_PILE') .. ' : ' .. spec.pile)
                    g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
                elseif actionName == InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_MOVE then
                    g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_AUTOLOADWOODEW_UNLOAD_MARKER_MOVE') .. ' : ' .. math.ceil(spec.unloadRot and spec.unloadPos or spec.unloadPos - 1) .. '/' .. math.ceil(spec.unloadRot and spec.rootShiftRotXMax or 4) .. ' m')
                    g_inputBinding:setActionEventTextVisibility(eventID, spec.showUnloadPos)
                elseif actionName == InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_SCALE then
                    g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_AUTOLOADWOODEW_UNLOAD_MARKER_SCALE') .. ' : ' .. math.ceil(spec.unloadScale * 4.8) .. '/' .. math.ceil(2 * 4.8) .. ' m')
                    g_inputBinding:setActionEventTextVisibility(eventID, spec.showUnloadPos)
                elseif actionName == InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_ROTATE then
                    g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_AUTOLOADWOODEW_UNLOAD_MARKER_ROTATE'))
                    g_inputBinding:setActionEventTextVisibility(eventID, spec.showUnloadPos and AutoLoadWoodCustom.showHelp)
                end
            end
            if spec.count > 0 then
                g_currentMission:addExtraPrintText(g_i18n:getText('AUTOLOADWOODEW_COUNT') .. " " .. spec.count .. " (" .. string.format('%.1f', spec.mass) .. " t)")
            else
                g_currentMission:addExtraPrintText(g_i18n:getText('AUTOLOADWOODEW_NO_LOGS'))
            end
        end
    end
end

function AutoLoadWoodCustom:onDraw()
	local spec = self.spec_autoLoadWoodCustom
end

function AutoLoadWoodCustom:toggleState(autoload, rightSide, unload, noEventSend)
	local spec = self.spec_autoLoadWoodCustom
	AutoLoadWoodCustomToggleEvent.sendEvent(self, autoload, rightSide, unload, noEventSend)
	if rightSide ~= spec.rightSide then
		if spec.unloadMarker then
			unlink(spec.unloadMarker)
			rotate(spec.unloadMarker, 0, math.rad(180), 0)
			local mX, mY, mZ = getTranslation(spec.unloadMarker)
			mX = mX * -1
			link(spec.unloadPoseRoot[rightSide and 2 or 1], spec.unloadMarker)
		end
	end
	spec.lastWoodShapes = 0
	spec.autoload = autoload
	spec.rightSide = rightSide
	spec.unload = unload
end

function AutoLoadWoodCustom:toggleLoad(showLoadPos, numPiles, pile, noEventSend)
	local spec = self.spec_autoLoadWoodCustom
	AutoLoadWoodCustomToggleLoadEvent.sendEvent(self, showLoadPos, numPiles, pile, noEventSend)
	spec.showLoadPos = showLoadPos
	spec.numPiles = numPiles
	spec.pile = pile
	self:updateValues()
	self:updateLoad()
	self:updateUnload()
end

function AutoLoadWoodCustom:toggleUnload(showUnloadPos, unloadPos, unloadScale, unloadRot, noEventSend)
	local spec = self.spec_autoLoadWoodCustom
	AutoLoadWoodCustomToggleUnloadEvent.sendEvent(self, showUnloadPos, unloadPos, unloadScale, unloadRot, noEventSend)
	if unloadPos ~= spec.unloadPos or unloadRot ~= spec.unloadRot then
		setTranslation(spec.unloadPoseNode[spec.rightSide and 2 or 1], spec.unloadPoseNodeX, spec.unloadPoseNodeY, spec.unloadPoseNodeZ)
	end
	spec.showUnloadPos = showUnloadPos
	spec.unloadPos = unloadPos
	spec.unloadScale = unloadScale
	spec.unloadRot = unloadRot
	self:updateValues()
	self:updateUnload()
end

function AutoLoadWoodCustom:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_autoLoadWoodCustom
	if keyStatus > 0 then
		if actionName == 'AUTOLOADWOODEW_TOGGLE_AUTO' then
			self:toggleState(not spec.autoload, spec.rightSide, false)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_UNLOAD' then
			self:toggleState(false, spec.rightSide, not spec.unload)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_SIDE' then
			self:toggleState(false, not spec.rightSide, false)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_PILES' then
            local maxPiles = spec.maxPiles or 4 -- Значение по умолчанию, если maxPiles nil
            if spec.numPiles + 1 <= maxPiles then
                spec.numPiles = spec.numPiles + 1
            else
                spec.numPiles = 1
            end
            self:toggleLoad(spec.showLoadPos, spec.numPiles, 1)
            self:toggleState(false, spec.rightSide, false)
            if spec.unloadRot then
                spec.unloadPos = spec.rootShiftRotX
            end
            self:toggleUnload(spec.showUnloadPos, spec.unloadPos, spec.unloadScale, spec.unloadRot)
		elseif actionName == 'AUTOLOADWOODEW_SWITCH_PILE' then
			if spec.pile + 1 <= spec.numPiles then
				spec.pile = spec.pile + 1
			else
				spec.pile = 1
			end
			self:toggleLoad(spec.showLoadPos, spec.numPiles, spec.pile)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_LOAD_MARKER' then
			self:toggleLoad(not spec.showLoadPos, spec.numPiles, spec.pile)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_UNLOAD_MARKER' then
			self:toggleUnload(not spec.showUnloadPos, spec.unloadPos, spec.unloadScale, spec.unloadRot)
		elseif actionName == 'AUTOLOADWOODEW_UNLOAD_MARKER_MOVE' then
			local rX, rY, rZ = getTranslation(spec.unloadPoseRoot[1])
			if spec.unloadRot == true then
				rX = math.ceil(math.abs(rX))
				if rX < spec.rootShiftRotXMax then
					rX = rX + 1
				else
					rX = spec.rootShiftRotX
				end
			else
				rX = math.ceil(math.abs(rX))
				if rX < 5 then
					rX = rX + 1
				else
					rX = spec.unloadPoseRootX
				end
			end
			local unloadPos = rX
			self:toggleUnload(spec.showUnloadPos, unloadPos, spec.unloadScale, spec.unloadRot)
		elseif actionName == 'AUTOLOADWOODEW_UNLOAD_MARKER_SCALE' then
			local msX, msY, msZ = getScale(spec.unloadMarker)
			if msZ < 2 then
				msZ = msZ + 0.2
			else
				msZ = 1
			end
			spec.unloadScale = msZ
			self:toggleUnload(spec.showUnloadPos, spec.unloadPos, spec.unloadScale, spec.unloadRot)
		elseif actionName == 'AUTOLOADWOODEW_UNLOAD_MARKER_ROTATE' then
			if spec.unloadRot == false then
				spec.unloadPos = spec.rootShiftRotX
			else
				spec.unloadPos = spec.unloadPoseRootX
			end
			self:toggleUnload(spec.showUnloadPos, spec.unloadPos, spec.unloadScale, not spec.unloadRot)
		elseif actionName == 'AUTOLOADWOODEW_TOGGLE_HELP' then
			AutoLoadWoodCustom.showHelp = not AutoLoadWoodCustom.showHelp
			self:saveToXml()
		end
	end
end

function AutoLoadWoodCustom:updateValues()
    local spec = self.spec_autoLoadWoodCustom
    print("Debug: updateValues called, trailer = " .. tostring(spec.trailer) .. ", numPiles = " .. tostring(spec.numPiles) .. ", maxPiles = " .. tostring(spec.maxPiles))

    -- Константы
    local minLengthPerPile = 2.5 -- Минимальная длина на штабель (включая space)
    local minLogLength = 2.0 -- Минимальная длина бревен
    local maxPilesLimit = 4 -- Максимальное количество штабелей
    local space = spec.space or 0.3 -- Расстояние между штабелями
    local maxLenFromXML = spec.maxLen or spec.trailer -- Максимальная длина бревен из XML

    -- Рассчитываем maxPiles, если не задано в XML
    if spec.maxPiles == nil then
        spec.maxPiles = math.min(math.floor(spec.trailer / minLengthPerPile), maxPilesLimit)
        print("Debug: maxPiles not set in XML, calculated maxPiles = " .. tostring(spec.maxPiles))
    else
        -- Ограничиваем maxPiles из XML разумными значениями
        spec.maxPiles = math.max(1, math.min(spec.maxPiles, maxPilesLimit))
        print("Debug: maxPiles set from XML = " .. tostring(spec.maxPiles))
    end

    -- Гарантируем минимум 1 штабель
    if spec.maxPiles < 1 then
        spec.maxPiles = 1
    end

    -- Корректируем numPiles, если оно больше maxPiles
    if spec.numPiles > spec.maxPiles then
        spec.numPiles = spec.maxPiles
    end

    -- Рассчитываем maxLength с учетом space
    spec.maxLength = (spec.trailer - (spec.numPiles - 1) * space) / spec.numPiles
    spec.maxLength = math.max(minLogLength, math.min(spec.maxLength, maxLenFromXML)) -- Ограничиваем min и max

    -- Рассчитываем startPosZ (центрируем штабели на грузовике)
    local totalLength = spec.numPiles * spec.maxLength + (spec.numPiles - 1) * space
    spec.startPosZ = -totalLength / 2 -- Смещаем к центру грузовика

    -- Позиция для конкретного pile
    if spec.numPiles > 1 then
        if spec.pile == 1 then
            spec.startPosZ = spec.startPosZ
        elseif spec.pile == 2 then
            spec.startPosZ = spec.startPosZ - spec.maxLength - space
        elseif spec.pile == 3 then
            spec.startPosZ = spec.startPosZ - 2 * (spec.maxLength + space)
        elseif spec.pile == 4 then
            spec.startPosZ = spec.startPosZ - 3 * (spec.maxLength + space)
        end
    end

    -- Рассчитываем rootShiftRotX и rootShiftRotXMax пропорционально длине
    local baseRotX = 9.65 * (spec.trailer / 15) -- Базовое значение для trailer=15
    local baseRotXMax = 12.8 * (spec.trailer / 15)
    spec.rootShiftRotX = baseRotX * (spec.maxLength / spec.trailer)
    spec.rootShiftRotXMax = baseRotXMax * (spec.maxLength / spec.trailer)

    -- Устанавливаем unloadShiftMax
    spec.unloadShiftMax = spec.unloadScale * 4

    print("Debug: updateValues finished, maxPiles = " .. tostring(spec.maxPiles) .. ", maxLength = " .. tostring(spec.maxLength) .. ", startPosZ = " .. tostring(spec.startPosZ))
end

function AutoLoadWoodCustom:updateLoad()
    local spec = self.spec_autoLoadWoodCustom
    if spec.fillPoseRoot then
        setTranslation(spec.fillPoseRoot, spec.fillPoseNodeX, spec.fillPoseNodeY, spec.startPosZ or 0)
    else
        print("Warning: fillPoseRoot is nil, check autoLoadWoodNode in XML")
    end
    if spec.loadMarker then
        setScale(spec.loadMarker, spec.maxLength or 1, 1, 2.5)
        setVisibility(spec.loadMarker, spec.showLoadPos)
    else
        print("Warning: loadMarker is nil, check loadMarker#node in XML")
    end
    if spec.unloadMarker then
        setScale(spec.unloadMarker, spec.maxLength or 1, 1, spec.unloadScale or 1)
    else
        print("Warning: unloadMarker is nil, check unloadMarker#node in XML")
    end
end

function AutoLoadWoodCustom:updateUnload()
    local spec = self.spec_autoLoadWoodCustom
    local j = 1
    for i = 1, 2 do
        if spec.unloadPoseRoot[i] then
            setRotation(spec.unloadPoseRoot[i], 0, (spec.unloadRot and math.rad(90) or math.rad(0)) * j, 0)
            setTranslation(spec.unloadPoseRoot[i], spec.unloadPos * j, 0, (spec.unloadRot and 3 or spec.unloadPoseRootZ))
        end
        j = j * -1
    end
    if spec.unloadMarker then
        setScale(spec.unloadMarker, spec.maxLength or 1, 1, spec.unloadScale or 1)
        setVisibility(spec.unloadMarker, spec.showUnloadPos)
    else
        print("Warning: unloadMarker is nil, check unloadMarker#node in XML")
    end
end

function AutoLoadWoodCustom:dynamicMountTriggerCallback(superFunc, triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	local spec = self.spec_autoLoadWoodCustom
	local splitType = g_splitShapeManager:getSplitTypeByIndex(getSplitType(otherActorId))
	if splitType ~= nil then
		if onEnter then
			if spec.woodLoad[otherActorId] == nil then
				spec.woodLoad[otherActorId] = otherActorId
			end
		elseif onLeave then
			if not spec.autoload then
				spec.woodLoad[otherActorId] = nil
			end
		end
	end
end

function AutoLoadWoodCustom:getIsTurnedOn()
	local spec = self.spec_autoLoadWoodCustom
	return spec.autoload or spec.unload
end

function AutoLoadWoodCustom:onReadStream(streamId, connection)
	local spec = self.spec_autoLoadWoodCustom
	local rightSide = streamReadBool(streamId)
	local showLoadPos = streamReadBool(streamId)
	local showUnloadPos = streamReadBool(streamId)
	local numPiles = streamReadInt8(streamId)
	local pile = streamReadInt8(streamId)
	local unloadPos = streamReadFloat32(streamId)
	local unloadScale = streamReadFloat32(streamId)
	local unloadRot = streamReadBool(streamId)
	if rightSide ~= nil and showLoadPos ~= nil and showUnloadPos ~= nil and numPiles ~= nil and pile ~= nil and unloadPos ~= nil and unloadScale ~= nil and unloadRot ~= nil then
		self:toggleState(spec.autoload, rightSide, spec.unload, true)
		self:toggleLoad(showLoadPos, numPiles, pile, true)
		self:toggleUnload(showUnloadPos, unloadPos, unloadScale, unloadRot, true)
	end
end

function AutoLoadWoodCustom:onWriteStream(streamId, connection)
	local spec = self.spec_autoLoadWoodCustom
	streamWriteBool(streamId, spec.rightSide)
	streamWriteBool(streamId, spec.showLoadPos)
	streamWriteBool(streamId, spec.showUnloadPos)
	streamWriteInt8(streamId, spec.numPiles)
	streamWriteInt8(streamId, spec.pile)
	streamWriteFloat32(streamId, spec.unloadPos)
	streamWriteFloat32(streamId, spec.unloadScale)
	streamWriteBool(streamId, spec.unloadRot)
end

--- Events ---

AutoLoadWoodCustomToggleEvent = {}
AutoLoadWoodCustomToggleEvent_mt = Class(AutoLoadWoodCustomToggleEvent, Event)
InitEventClass(AutoLoadWoodCustomToggleEvent, "AutoLoadWoodCustomToggleEvent")

function AutoLoadWoodCustomToggleEvent.emptyNew()
	local self = Event.new(AutoLoadWoodCustomToggleEvent_mt)
	return self
end

function AutoLoadWoodCustomToggleEvent.new(object, autoload, rightSide, unload)
	local self = AutoLoadWoodCustomToggleEvent.emptyNew()
	self.object = object
	self.autoload = autoload
	self.rightSide = rightSide
	self.unload = unload
	return self
end

function AutoLoadWoodCustomToggleEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.autoload = streamReadBool(streamId)
	self.rightSide = streamReadBool(streamId)
	self.unload = streamReadBool(streamId)
	self:run(connection)
end

function AutoLoadWoodCustomToggleEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.autoload)
	streamWriteBool(streamId, self.rightSide)
	streamWriteBool(streamId, self.unload)
end

function AutoLoadWoodCustomToggleEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end
	if self.object ~= nil then
		self.object:toggleState(self.autoload, self.rightSide, self.unload, true)
	end
end

function AutoLoadWoodCustomToggleEvent.sendEvent(vehicle, autoload, rightSide, unload, noEventSend)
	if autoload ~= vehicle.autoload or rightSide ~= vehicle.rightSide or unload ~= vehicle.unload then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(AutoLoadWoodCustomToggleEvent.new(vehicle, autoload, rightSide, unload), nil, nil, vehicle)
			else
				g_client:getServerConnection():sendEvent(AutoLoadWoodCustomToggleEvent.new(vehicle, autoload, rightSide, unload))
			end
		end
	end
end

AutoLoadWoodCustomToggleLoadEvent = {}
AutoLoadWoodCustomToggleLoadEvent_mt = Class(AutoLoadWoodCustomToggleLoadEvent, Event)
InitEventClass(AutoLoadWoodCustomToggleLoadEvent, "AutoLoadWoodCustomToggleLoadEvent")

function AutoLoadWoodCustomToggleLoadEvent.emptyNew()
	local self = Event.new(AutoLoadWoodCustomToggleLoadEvent_mt)
	return self
end

function AutoLoadWoodCustomToggleLoadEvent.new(object, showLoadPos, numPiles, pile)
	local self = AutoLoadWoodCustomToggleLoadEvent.emptyNew()
	self.object = object
	self.showLoadPos = showLoadPos
	self.numPiles = numPiles
	self.pile = pile
	return self
end

function AutoLoadWoodCustomToggleLoadEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.showLoadPos = streamReadBool(streamId)
	self.numPiles = streamReadInt8(streamId)
	self.pile = streamReadInt8(streamId)
	self:run(connection)
end

function AutoLoadWoodCustomToggleLoadEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.showLoadPos)
	streamWriteInt8(streamId, self.numPiles)
	streamWriteInt8(streamId, self.pile)
end

function AutoLoadWoodCustomToggleLoadEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end
	if self.object ~= nil then
		self.object:toggleLoad(self.showLoadPos, self.numPiles, self.pile, true)
	end
end

function AutoLoadWoodCustomToggleLoadEvent.sendEvent(vehicle, showLoadPos, numPiles, pile, noEventSend)
	if showLoadPos ~= vehicle.showLoadPos or numPiles ~= vehicle.numPiles or pile ~= vehicle.pile then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(AutoLoadWoodCustomToggleLoadEvent.new(vehicle, showLoadPos, numPiles, pile), nil, nil, vehicle)
			else
				g_client:getServerConnection():sendEvent(AutoLoadWoodCustomToggleLoadEvent.new(vehicle, showLoadPos, numPiles, pile))
			end
		end
	end
end

AutoLoadWoodCustomToggleUnloadEvent = {}
AutoLoadWoodCustomToggleUnloadEvent_mt = Class(AutoLoadWoodCustomToggleUnloadEvent, Event)
InitEventClass(AutoLoadWoodCustomToggleUnloadEvent, "AutoLoadWoodCustomToggleUnloadEvent")

function AutoLoadWoodCustomToggleUnloadEvent.emptyNew()
	local self = Event.new(AutoLoadWoodCustomToggleUnloadEvent_mt)
	return self
end

function AutoLoadWoodCustomToggleUnloadEvent.new(object, showUnloadPos, unloadPos, unloadScale, unloadRot)
	local self = AutoLoadWoodCustomToggleUnloadEvent.emptyNew()
	self.object = object
	self.showUnloadPos = showUnloadPos
	self.unloadPos = unloadPos
	self.unloadScale = unloadScale
	self.unloadRot = unloadRot
	return self
end

function AutoLoadWoodCustomToggleUnloadEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.showUnloadPos = streamReadBool(streamId)
	self.unloadPos = streamReadFloat32(streamId)
	self.unloadScale = streamReadFloat32(streamId)
	self.unloadRot = streamReadBool(streamId)
	self:run(connection)
end

function AutoLoadWoodCustomToggleUnloadEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.showUnloadPos)
	streamWriteFloat32(streamId, self.unloadPos)
	streamWriteFloat32(streamId, self.unloadScale)
	streamWriteBool(streamId, self.unloadRot)
end

function AutoLoadWoodCustomToggleUnloadEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end
	if self.object ~= nil then
		self.object:toggleUnload(self.showUnloadPos, self.unloadPos, self.unloadScale, self.unloadRot, true)
	end
end

function AutoLoadWoodCustomToggleUnloadEvent.sendEvent(vehicle, showUnloadPos, unloadPos, unloadScale, unloadRot, noEventSend)
	if showUnloadPos ~= vehicle.showUnloadPos or unloadPos ~= vehicle.unloadPos or unloadScale ~= vehicle.unloadScale or unloadRot ~= vehicle.unloadRot then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(AutoLoadWoodCustomToggleUnloadEvent.new(vehicle, showUnloadPos, unloadPos, unloadScale, unloadRot), nil, nil, vehicle)
			else
				g_client:getServerConnection():sendEvent(AutoLoadWoodCustomToggleUnloadEvent.new(vehicle, showUnloadPos, unloadPos, unloadScale, unloadRot))
			end
		end
	end
end

function AutoLoadWoodCustom:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	local spec = self.spec_autoLoadWoodCustom
	if g_dedicatedServerInfo ~= nil then
		return
	end
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if isSelected then
		local actions_RC1 = {
			InputAction.AUTOLOADWOODEW_TOGGLE_AUTO,
			InputAction.AUTOLOADWOODEW_TOGGLE_UNLOAD,
			InputAction.AUTOLOADWOODEW_TOGGLE_SIDE,
			InputAction.AUTOLOADWOODEW_TOGGLE_PILES,
			InputAction.AUTOLOADWOODEW_SWITCH_PILE,
			InputAction.AUTOLOADWOODEW_TOGGLE_LOAD_MARKER,
			InputAction.AUTOLOADWOODEW_TOGGLE_UNLOAD_MARKER,
			InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_MOVE,
			InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_SCALE,
			InputAction.AUTOLOADWOODEW_UNLOAD_MARKER_ROTATE,
			InputAction.AUTOLOADWOODEW_TOGGLE_HELP
		}
		for _, actionName in pairs(actions_RC1) do
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, AutoLoadWoodCustom.actionCallback, true, true, false, true)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == InputAction.AUTOLOADWOODEW_TOGGLE_AUTO or actionName == InputAction.AUTOLOADWOODEW_TOGGLE_UNLOAD or actionName == InputAction.AUTOLOADWOODEW_TOGGLE_SIDE then
					g_inputBinding.events[eventID].displayPriority = 1
				else
					g_inputBinding.events[eventID].displayPriority = 2
				end
				if actionName == InputAction.AUTOLOADWOODEW_TOGGLE_HELP then
					g_inputBinding:setActionEventTextVisibility(eventID, true)
				elseif actionName == InputAction.RC_TOGGLE or actionName == InputAction.RC_TOGGLE_ALL then
					g_inputBinding:setActionEventTextVisibility(eventID, false)
				else
					g_inputBinding:setActionEventTextVisibility(eventID, AutoLoadWoodCustom.showHelp)
				end
			end
		end
	end
end